/* $VER: vlink ldscript.c V0.8 (27.01.01)
 *
 * This file is part of vlink, a portable linker for multiple
 * object formats.
 * Copyright (c) 1997-2001  Frank Wille
 *
 * vlink is freeware and part of the portable and retargetable ANSI C
 * compiler vbcc, copyright (c) 1995-2001 by Volker Barthelmann.
 * vlink may be freely redistributed as long as no modifications are
 * made and nothing is charged for it. Non-commercial usage is allowed
 * without any restrictions.
 * EVERY PRODUCT OR PROGRAM DERIVED DIRECTLY FROM MY SOURCE MAY NOT BE
 * SOLD COMMERCIALLY WITHOUT PERMISSION FROM THE AUTHOR.
 *
 *
 * v0.8  (27.01.00) phx
 *       The way to call addglobsym(), to make a symbol of global scope,
 *       has changed.
 * v0.6e (09.10.99) phx
 *       File created.
 */


#define LDSCRIPT_C
#include "vlink.h"
#include "ldscript.h"

#define DEFLEN 0x7fffffff  /* default length for new memory blocks */

static char *current_addr = ".";  /* symbol represents current address */
static char *scriptname;
static char *sptr;  /* current script source pointer */
static int line;    /* current line number */
static struct MemoryDescr *defmem;  /* current default memory block */

static struct LdFuncs ldfuncs[] = {
  { "ALIGN",ldfunc_align },
  { NULL,NULL }
};



void read_ld_script(struct GlobalVars *gv,const char *script)
{
  if (script) {
    struct LdScript ldscript;
    struct LdScript *lds = &ldscript;
    char *keyword;
    struct LinkedSection *ls;

    /* initialize LdScript structure */
    lds->fff = fff[gv->dest_format];
    lds->alignment = 2;  /* default alignment = 32bit */
    initmemblks(lds);
    initldsyms(lds);
    initlist(&lds->early_symbols);

    /* parse script */
    scriptname = gv->scriptname;
    sptr = (char *)script;
    line = 1;
    defmem = getmemblk(lds,"default");

    while (keyword = getword()) {
      char c = getchr();

      if (c == '{') {
        if (!strcmp(keyword,"MEMORY"))
          define_memory(lds);
        else if (!strcmp(keyword,"SECTIONS"))
          define_sections(gv,lds);
        else {
          error(65,scriptname,line,keyword);  /* unknown keyword ignored */
          skipblock(1,'{','}');
        }
      }

      else if (c == '(')
        dofunction(gv,lds,keyword);

      else
        error(65,scriptname,line,keyword);  /* unknown keyword ignored */
    }

    gv->use_ldscript = TRUE;
  }
}


static void define_memory(struct LdScript *lds)
/* Syntax: */
/* <memblockname>: o[rg] = <base address>, l[en] = <maxlen> */
{
  char memname[MAXLEN];
  char *str;
  unsigned long org,len;

  while (str = getword()) {
    strcpy(memname,str);
    if (getchr() == ':') {
      org = readmemassign("org");
      if (getchr() == ',') {
        len = readmemassign("len");
        addmemblk(lds,memname,org,len);
      }
      else
        error(66,scriptname,line,',');  /* ',' expected */
    }
    else
      error(66,scriptname,line,':');    /* ':' expected */
  }
  endofblock();
}


static void defmemblock(struct LdScript *lds,unsigned long val)
{
  char namebuf[MAXLEN];

  sprintf(namebuf,"mem0x%08lx",val);
  defmem = addmemblk(lds,namebuf,val,DEFLEN);
  if (!listempty(&lds->early_symbols)) {
    error(73,scriptname,line);  /* can't def. sym. without a sect. */
    initlist(&lds->early_symbols);
  }
}


static void define_sections(struct GlobalVars *gv,struct LdScript *lds)
/* Syntax: */
/* <section>: { <file-pattern>(<sect.-pattern>) ... } [> <memblockname>] */
/* <symbol> = <expression>; */
{
  char namebuf[MAXLEN];
  struct LinkedSection *ls = NULL;
  struct MemoryDescr *md;
  char *str;

  while (str = getword()) {
    switch (getchr()) {

      case '=':   /* symbol definition */
        ls = symbol_assign(gv,lds,ls,str,TRUE);
        break;

      case '(':   /* function call */
        dofunction(gv,lds,str);
        break;

      case ':':   /* section definition */
        ls = section_block(gv,lds,NULL,str);
        break;

      case '@':   /* section definition at different memory block */
        strcpy(namebuf,str);
        if (str = getword()) {
          if (!(md = getmemblk(lds,str))) {
            error(70,scriptname,line,str);  /* unknown memory description */
            break;
          }
        }
        else {
          error(71,scriptname,line);  /* missing memory descr. specifier */
          break;
        }
        if (getchr() == ':') {  /* section definition */
          ls = section_block(gv,lds,md,str);
          break;
        }
        /* fall through to syntax error */

      default:
        error(72,scriptname,line);  /* syntax error */
        break;
    }
  }
  endofblock();
}


static void dofunction(struct GlobalVars *gv,struct LdScript *lds,
                       char *name)
{
  char funcname[MAXLEN];
  char *arg;

  strcpy(funcname,name);
  if (arg = getalnum()) {
    /* find function in LdFuncs list */
    struct LdFuncs *f = ldfuncs;

    while (f->name) {
      if (!strcmp(f->name,funcname)) {
        f->funcptr(gv,lds,arg);  /* call function */
        break;
      }
      f++;
    }
    if (f->name == NULL)
      error(80,scriptname,line,funcname);  /* unknown linker script function */
  }
  else
    error(78,scriptname,line);  /* missing argument */
  endoffunc();
}


static void ldfunc_align(struct GlobalVars *gv,struct LdScript *lds,
                         char *arg)
/* ALIGN() : set alignment value for following sections */
{
  lds->alignment = (uint8)readval(arg);
}


static struct LinkedSection *section_block(struct GlobalVars *gv,
                                           struct LdScript *lds,
                                           struct MemoryDescr *md,
                                           char *sectname)
/* <section>: { <file-pattern>(<sect.-pattern>) ... } [> <memblockname>] */
{
  char namebuf[MAXLEN];
  char *str,ch;
  struct LinkedSection *ls = NULL;
  struct Symbol *sym;
  bool endsymdef = FALSE;  /* symbol at end of section already defined */

  if (getchr() == '{') {
    while (str = getpattern()) {
      strcpy(namebuf,str);
      ch = getchr();

      if (ch == '(') {
        /* file-/section-pattern definition */
        if (str = getpattern()) {
          if (getchr() == ')') {
            if (ls == NULL)
              ls = create_lnksect(gv,allocstring(sectname),0,0,0,
                                  lds->alignment);
            addpattern(ls,namebuf,str);
            if (endsymdef)
              error(82,scriptname,line);  /* no symbols within a section */
          }
          else
            error(66,scriptname,line,')');  /* ')' expected */
        }
        else
          error(69,scriptname,line);  /* missing section pattern */
      }

      else if (ch == '=') {
        /* symbol assignment before or after section */
        symbol_assign(gv,lds,ls,str,FALSE);
        if (ls)
          endsymdef = TRUE;  /* there should not be another sect.pattern */
      }

      else
        error(72,scriptname,line);  /* syntax error */
    }
    endofblock();

    /* assign relocation memory block */
    if (getchr() == '>') {  /* non-default memory block specified? */
      if (str = getword()) {
        if (ls->relocmem = getmemblk(lds,str)) {
          defmem = ls->relocmem;
        }
        else {
          error(70,scriptname,line,str);  /* unknown memory description */
          ls->relocmem = defmem;
        }
      }
      else {
        error(71,scriptname,line);  /* missing memory descr. specifier */
        ls->relocmem = defmem;
      }
    }
    else {
      --sptr;
      ls->relocmem = defmem;
    }

    /* assign destination memory block */
    if (md)
      ls->destmem = md;
    else
      ls->destmem = defmem;

    /* assign early-symbols, which were defined before this section */
    while (sym = (struct Symbol *)remhead(&lds->early_symbols))
      addtail(&ls->scriptsymbols,&sym->n);
  }
  else
    error(66,scriptname,line,'{');  /* '{' expected */

  return (ls);
}


static struct LinkedSection *symbol_assign(struct GlobalVars *gv,
                                           struct LdScript *lds,
                                           struct LinkedSection *ls,
                                           char *symname,bool newaddr)
/* Evaluate "<symbol> = <expression>;" in a SECTION block. */
/* The '=' was already read, so the stream-pointer is ready for evalexp(). */
{
  char namebuf[MAXLEN];
  unsigned long val;
  bool absolute;  /* true: val is absolute, false: offset to current addr. */

  if (!strcmp(symname,current_addr)) {      /* . = <exp> */
    if (newaddr) {
      if (evalexp(lds,&val,&absolute)) {
        if (absolute) {  /* new address requires new MemBlock */
          defmemblock(lds,val);
          ls = NULL;  /* last section is invalid now */
        }
        else {  /* add an offset to current address */
          if ((long)val >= 0) {
            if (ls) {
              ls->bsssize += val;  /* append 0-bytes */
            }
            else {  /* @@@ hmm... */
              defmem->org += val;
              defmem->current = defmem->org;
              defmem->len -= val;
            }    
          }
          else
            error(79,scriptname,line); /* sub. from cur. addr. not allowed */
        }
      }
    }
    else
      error(81,scriptname,line);  /* can't change addr. within a section */
  }

  else {                                    /* <symbol> = <exp> */
    strcpy(namebuf,symname);
    if (evalexp(lds,&val,&absolute)) {
      addscriptsymbol(gv,lds,ls,namebuf,val,absolute?SYM_ABS:SYM_RELOC,
                      absolute?0:(ls?SYMF_ENDOFSECTREL:0));
    }
  }

  return (ls);
}


static bool evalexp(struct LdScript *lds,unsigned long *val,bool *absolute)
/* evaluate a linker-script expression, which is terminated by a ';' */
{
  char *arg,c;
  long av;
  bool minus = FALSE;
  struct LdSymbol *ldsym;

  *val = 0;
  *absolute = TRUE;

  while (arg = getalnum()) {

    if (isdigit((unsigned)*arg)) {
      /* absolute numerical constant */
      av = readval(arg);
      if (minus)
        av = -av;
      *val += (unsigned long)av;
    }

    else if (!strcmp(current_addr,arg)) {
      /* '.' address symbol */
      if (!*absolute) {
        error(74,scriptname,line);  /* double '.' in expression */
        return (FALSE);
      }
      if (minus) {
        error(75,scriptname,line);  /* can't subtract an address */
        return (FALSE);
      }
      *absolute = FALSE;
    }

    else {
      /* linker script symbol */
      ldsym = lds->special_symbols;
      while (ldsym) {
        if (!strcmp(ldsym->name,arg))
          break;
        ldsym = ldsym->next;
      }
      if (ldsym) {
        if (minus)
          *val -= ldsym->value;
        else
          *val += ldsym->value;
      }
      else {
        error(76,scriptname,line);  /* reference to unknown symbol */
        return (FALSE);
      }
    }

    c = getchr();
    if (c == ';')  /* expression ends here */
      break;
    else if (c == '+')
      minus = FALSE;
    else if (c == '-')
      minus = TRUE;
    else {
      sptr--;
      error(77,scriptname,line,c);  /* illegal operation */
      return (FALSE);
    }
  }

  return (TRUE);
}


static unsigned long readmemassign(char *key)
{
  char *str;
  long val = 0;

  if (str = getword()) {
    if (*str == *key) {
      if (getchr() == '=') {
        if (str = getnum())
          val = readval(str);
        else
          error(67,scriptname,line);  /* absolute number expected */
      }
      else
        error(66,scriptname,line,'=');
    }
    else
      error(68,scriptname,line,key);  /* keyword expected */
  }
  else
    error(68,scriptname,line,key);
  return ((unsigned long)val);
}


static void endofblock(void)
{
  if (getchr() != '}') {
    error(66,scriptname,line,'}');  /* '}' expected */
    skipblock(1,'{','}');
  }
}


static void endoffunc(void)
{
  if (getchr() != ')') {
    error(66,scriptname,line,')');  /* ')' expected */
    skipblock(1,'(',')');
  }
}


static void initldsyms(struct LdScript *lds)
/* initialize special symbols, which can be used in a linker-script */
{
  lds->special_symbols = NULL;
  addldsym(lds,"SIZEOF_HEADERS",lds->fff->headersize);
}


static void initmemblks(struct LdScript *lds)
/* set default memory block */
{
  lds->memory_blocks = NULL;
  addmemblk(lds,"default",0,DEFLEN);
}


static void addldsym(struct LdScript *lds,char *name,unsigned long value)
{
  struct LdSymbol *lastsym,*sym = alloc(sizeof(struct LdSymbol));

  if (lastsym = lds->special_symbols) {
    while (lastsym->next)
      lastsym = lastsym->next;
    lastsym->next = sym;
  }
  else
    lds->special_symbols = sym;

  sym->next = NULL;
  sym->name = allocstring(name);
  sym->value = value;
}


static struct MemoryDescr *addmemblk(struct LdScript *lds,char *name,
                                     unsigned long org,unsigned long len)
{
  struct MemoryDescr *last,*new = alloc(sizeof(struct MemoryDescr));

  if (last = lds->memory_blocks) {
    while (last->next)
      last = last->next;
    last->next = new;
  }
  else
    lds->memory_blocks = new;

  new->next = NULL;
  new->name = allocstring(name);
  new->org = new->current = org;
  if (org+len < org)
    len -= org+len+1;
  new->len = len;
  return (new);
}


static struct MemoryDescr *getmemblk(struct LdScript *lds,char *name)
{
  struct MemoryDescr *md = lds->memory_blocks;

  while (md) {
    if (!strcmp(md->name,name))
      return (md);
    md = md->next;
  }
  return (NULL);
}


static void addpattern(struct LinkedSection *ls,char *fname,char *sname)
{
  struct SectPattern *last,*new = alloc(sizeof(struct SectPattern));

  if (last = ls->pattern) {
    while (last->next)
      last = last->next;
    last->next = new;
  }
  else
    ls->pattern = new;

  new->next = NULL;
  new->filename = allocstring(fname);
  new->sectname = allocstring(sname);
}


static void addscriptsymbol(struct GlobalVars *gv,struct LdScript *lds,
                            struct LinkedSection *ls,char *name,
                            uint32 val,uint8 type,uint8 flags)
{
  struct Symbol *sym = alloc(sizeof(struct Symbol));

  sym->glob_chain = sym->obj_chain = NULL;
  sym->name = allocstring(name);
  sym->value = val;
  sym->relsect = NULL;
  sym->type = type;
  sym->flags = flags;
  sym->info = SYMI_OBJECT;
  sym->bind = SYMB_GLOBAL;
  sym->size = sym->extra = 0;

  if (addglobsym(gv,sym)) {
    if (ls)
      addtail(&ls->scriptsymbols,&sym->n);
    else
      /* sym. will be assigned to the next, still undefd., section */
      addtail(&lds->early_symbols,&sym->n);
  }
  else
    free(sym);
}


static char *getword(void)
{
  return (getarg(1));
}


static char *getalnum(void)
{
  return (getarg(3));
}


static char *getnum(void)
{
  char *num = getarg(3);

  if (!isdigit((unsigned)*num))
    return (NULL);
  return (num);
}


static char *getpattern(void)
{
  return (getarg(5));
}


static char *getarg(uint8 mask)
{
/* table of valid characters, */
/* 0=invalid, 1=valid in whole word, 2=valid, but not as first char */
/* 4=symbols for pattern-matching */
  static uint8 validchars[256] = {
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,4,0,0,0,1,0,  /* . */
    2,2,2,2,2,2,2,2,2,2,0,0,0,0,0,4,  /* 0-9 */
    0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,  /* A-O */
    1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,  /* P-Z _ */
    0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,  /* a-o */
    1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,  /* p-z */
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
  };
  static char buffer[MAXLEN];
  char *bp = buffer;
  char *sp = skip(sptr);

  /* try to read next word into buffer */
  while (validchars[(unsigned char)*sp] & mask) {
    mask |= 2;
    *bp++ = *sp++;
    if (bp >= &buffer[MAXLEN-1])
      break;  /* buffer overflow */
  }
  *bp = '\0';
  sptr = sp;

  if (buffer[0])
    return (buffer);
  return (NULL);  /* no valid word read */
}


static char getchr(void)
{
  char *sp = skip(sptr);
  char c;

  if (c = *sp)
    sp++;
  sptr = sp;
  return (c);
}


static void skipblock(int level,char start,char end)
/* skips a block between the two specified start- and end-characters */
{
  char c;

  while (c = getchr()) {
    if (c == start)
      level++;
    else if (c == end) {
      if (--level <= 0)
        break;
    }
  }
}


static char *skip(char *sp)
/* skips blanks, tabs, newlines and comments */
{
  sp = skipblanks(sp);
  while (*sp == '/' && *(sp+1) == '*') {
    /* skip comment */
    sp += 2;
    while (*sp && (*sp!='*' || *(sp+1)!='/')) {
      if (*sp=='\n')
        line++;
      sp++;
    }
    sp += 2;
    sp = skipblanks(sp);
  }
  return (sp);
}


static char *skipblanks(char *sp)
{
  char c = *sp;

  while (c==' ' || c=='\t' || c=='\n' || c=='\r') {
    if (c == '\n')
      line++;
    c = *(++sp);
  }
  return (sp);
}


static long readval(char *arg)
{
  long val;

  sscanf(arg,"%li",&val);
  return (val);
}
